import os

import uvicorn
from fastapi import FastAPI
from watchfiles import Change, DefaultFilter, run_process

from steroid.controller import Controller
from steroid.logging import getLogger
from steroid.utils import clearTerminal, removeLeadingOrTrailingSlash


class OnlyPythonFilesFilter(DefaultFilter):
    allowed_extensions = ".py"

    def __call__(self, change: Change, path: str) -> bool:
        return super().__call__(change, path) and path.endswith(self.allowed_extensions)


class CreateApp:
    _APP: FastAPI = None
    logger = getLogger()

    def __new__(cls, *args, **kwargs):
        if cls._APP is None:
            cls._APP = FastAPI(*args, **kwargs)

        return cls

    @classmethod
    @property
    def app(cls):
        return cls._APP

    @classmethod
    def _start(cls):
        clearTerminal()

        uvicorn.run(cls.app)

    @classmethod
    def onFileChange(cls, change: Change):
        getLogger().info(f"reloading ... file changes in {change}")

    @classmethod
    def start(cls, reload=True):
        if reload == False:
            cls._start()
            return

        watchFilePath = os.getcwd()

        run_process(
            watchFilePath,
            target=cls._start,
            callback=cls.onFileChange,
            watch_filter=OnlyPythonFilesFilter(),
        )

    @classmethod
    def addController(cls, controller: Controller):
        cls.app.include_router(
            controller.router,
            prefix=controller.path,
            tags=[removeLeadingOrTrailingSlash(controller.path)]
            if controller.path
            else [],
        )
        cls.logger.info(f"Mapped {controller.path} to router")



import inspect

from steroid.utils import formatPath


class BaseMethod:
    _METHOD: str

    def __init__(self, path: str = "", *args, **kwargs):
        self.path = formatPath(path)

        self.args = args
        self.kwargs = kwargs

        self.route = None

    def __call__(self, func, router=None):
        # TODO: alot feels very wrong in this method, help me!

        self.func = func

        if router:
            httpMethodDecorator = getattr(router, self._METHOD.lower())

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            parameters = [
                parameter
                for parameter in dict(inspect.signature(func).parameters).values()
            ]

            httpMethodDecorator(self.path)(func)(*parameters)

            return wrapper

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        setattr(wrapper, "object", self)

        return wrapper

    def mapSelfToRouter(self, router):
        self.__call__(self.func, router)


class Get(BaseMethod):
    _METHOD = "GET"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Post(BaseMethod):
    _METHOD = "POST"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Put(BaseMethod):
    _METHOD = "PUT"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Patch(BaseMethod):
    _METHOD = "PATCH"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Delete(BaseMethod):
    _METHOD = "DELETE"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import logging
from logging.config import dictConfig

from pydantic import BaseModel


class LogConfig(BaseModel):
    """Logging configuration to be set for the server"""

    LOGGER_NAME: str = "mycoolapp"
    LOG_FORMAT: str = "%(levelprefix)s %(message)s"
    LOG_LEVEL: str = "DEBUG"

    # Logging config
    version: int = 1
    disable_existing_loggers: bool = False
    formatters: dict = {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": LOG_FORMAT,
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    }
    handlers: dict = {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
    }
    loggers: dict = {
        LOGGER_NAME: {"handlers": ["default"], "level": LOG_LEVEL},
    }


def getLogger():
    dictConfig(LogConfig().model_dump())
    logger = logging.getLogger("mycoolapp")

    return logger



from fastapi import HTTPException, Request
from fastapi.responses import JSONResponse

from steroid.utils import getApp


class Middleware:
    HOOK: str = "PRE"

    def __call__(self, cls):
        return cls

    def action(request: Request):
        raise NotImplementedError("Middleware action must be implemented")


class UseMiddlewares:
    _PRE_MIDDLEWARES: list[Middleware] = []
    _POST_MIDDLEWARES: list[Middleware] = []

    def __init__(self, *middlewares: list[Middleware]):
        pre_middlewares = []
        post_middlewares = []

        for middleware in middlewares:
            if middleware.HOOK == "PRE":
                pre_middlewares.append(middleware)
            elif middleware.HOOK == "POST":
                post_middlewares.append(middleware)

        self._PRE_MIDDLEWARES = pre_middlewares
        self._POST_MIDDLEWARES = post_middlewares

    def __call__(self, cls):
        applyMiddlewareToRoute = getattr(cls, "path")

        async def routeMiddlewareAction(request: Request, call_next):
            try:
                canApplyToRoute = request.url.path.startswith(applyMiddlewareToRoute)

                if canApplyToRoute:
                    for middleware in self._PRE_MIDDLEWARES:
                        middleware.action(request)

                response = await call_next(request)

                if canApplyToRoute:
                    for middleware in self._POST_MIDDLEWARES:
                        middleware.action(request)

                return response
            except HTTPException as error:
                return JSONResponse(
                    status_code=error.status_code, content={"detail": error.detail}
                )

        rootApp = getApp()
        rootApp.middleware("http")(routeMiddlewareAction)

        return cls



from typing import Any

from fastapi.routing import APIRouter

from steroid.methods import BaseMethod
from steroid.utils import formatPath


class Controller:
    def __init__(self, path: str = "", *args, **kwargs):
        self.router = APIRouter()

        self.path = formatPath(path)

        self.args = args
        self.kwargs = kwargs

    def __call__(self, cls):
        self.setupController(cls)

        methods = list(self.findMethods(cls))
        self.mapMethodsToController(methods)

        return cls(*self.args, **self.kwargs)

    def setupController(self, classObject):
        setattr(classObject, "path", self.path)
        setattr(classObject, "router", self.router)

    def findMethods(self, classObject):
        for _, method in classObject.__dict__.items():
            methodClass = getattr(method, "object", None)
            if isinstance(methodClass, BaseMethod):
                yield methodClass

    def mapMethodsToController(self, methods: list[BaseMethod]):
        return list(
            map(lambda method: method.mapSelfToRouter(router=self.router), methods)
        )
